@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2012 Brazil
+  Copyright(C) 2012-2013 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -34,10 +34,17 @@ typedef struct {
   char *database_path_cstr;
   ngx_flag_t database_auto_create;
   ngx_str_t base_path;
+  ngx_str_t log_path;
+  ngx_open_file_t *log_file;
+  grn_log_level log_level;
+  ngx_str_t query_log_path;
+  ngx_open_file_t *query_log_file;
+  size_t cache_limit;
   char *config_file;
   int config_line;
   char *name;
   grn_ctx context;
+  grn_cache *cache;
 } ngx_http_groonga_loc_conf_t;
 
 typedef struct {
@@ -47,12 +54,24 @@ typedef struct {
 } ngx_http_groonga_database_callback_data_t;
 
 typedef struct {
+  grn_bool initialized;
   grn_ctx context;
   grn_obj head;
   grn_obj body;
   grn_obj foot;
 } ngx_http_groonga_handler_data_t;
 
+typedef struct {
+  ngx_pool_t *pool;
+  ngx_open_file_t *file;
+} ngx_http_groonga_logger_data_t;
+
+typedef struct {
+  ngx_pool_t *pool;
+  ngx_open_file_t *file;
+  ngx_str_t *path;
+} ngx_http_groonga_query_logger_data_t;
+
 typedef void (*ngx_http_groonga_loc_conf_callback_pt)(ngx_http_groonga_loc_conf_t *conf, void *user_data);
 
 ngx_module_t ngx_http_groonga_module;
@@ -83,6 +102,192 @@ ngx_str_equal_c_string(ngx_str_t *string, const char *c_string)
   return memcmp(c_string, string->data, string->len) == 0;
 }
 
+static void
+ngx_http_groonga_logger_log(grn_ctx *ctx, grn_log_level level,
+                            const char *timestamp, const char *title,
+                            const char *message, const char *location,
+                            void *user_data)
+{
+  ngx_http_groonga_logger_data_t *logger_data = user_data;
+  const char level_marks[] = " EACewnid-";
+  u_char buffer[NGX_MAX_ERROR_STR];
+  u_char *last;
+
+  if (location && *location) {
+    last = ngx_slprintf(buffer, buffer + NGX_MAX_ERROR_STR,
+                        "%s|%c|%s %s %s\n",
+                        timestamp, *(level_marks + level), title, message,
+                        location);
+  } else {
+    last = ngx_slprintf(buffer, buffer + NGX_MAX_ERROR_STR,
+                        "%s|%c|%s %s\n",
+                        timestamp, *(level_marks + level), title, message);
+  }
+  ngx_write_fd(logger_data->file->fd, buffer, last - buffer);
+}
+
+static void
+ngx_http_groonga_logger_reopen(grn_ctx *ctx, void *user_data)
+{
+  GRN_LOG(ctx, GRN_LOG_NOTICE, "log will be closed.");
+  ngx_reopen_files((ngx_cycle_t *)ngx_cycle, -1);
+  GRN_LOG(ctx, GRN_LOG_NOTICE, "log opened.");
+}
+
+static void
+ngx_http_groonga_logger_fin(grn_ctx *ctx, void *user_data)
+{
+  ngx_http_groonga_logger_data_t *logger_data = user_data;
+
+  ngx_pfree(logger_data->pool, logger_data);
+}
+
+static grn_logger ngx_http_groonga_logger = {
+  GRN_LOG_DEFAULT_LEVEL,
+  GRN_LOG_TIME | GRN_LOG_MESSAGE,
+  NULL,
+  ngx_http_groonga_logger_log,
+  ngx_http_groonga_logger_reopen,
+  ngx_http_groonga_logger_fin
+};
+
+static ngx_int_t
+ngx_http_groonga_context_init_logger(grn_ctx *context,
+                                     ngx_http_groonga_loc_conf_t *location_conf,
+                                     ngx_pool_t *pool,
+                                     ngx_log_t *log)
+{
+  ngx_http_groonga_logger_data_t *logger_data;
+
+  if (!location_conf->log_file) {
+    return NGX_OK;
+  }
+
+  logger_data = ngx_pcalloc(pool, sizeof(ngx_http_groonga_logger_data_t));
+  if (!logger_data) {
+    ngx_log_error(NGX_LOG_ERR, log, 0,
+                  "http_groonga: failed to allocate memory for logger");
+    return NGX_ERROR;
+  }
+
+  logger_data->pool = pool;
+  logger_data->file = location_conf->log_file;
+  ngx_http_groonga_logger.max_level = location_conf->log_level;
+  ngx_http_groonga_logger.user_data = logger_data;
+  grn_logger_set(context, &ngx_http_groonga_logger);
+
+  return NGX_OK;
+}
+
+static void
+ngx_http_groonga_query_logger_log(grn_ctx *ctx, unsigned int flag,
+                                  const char *timestamp, const char *info,
+                                  const char *message, void *user_data)
+{
+  ngx_http_groonga_query_logger_data_t *data = user_data;
+  u_char buffer[NGX_MAX_ERROR_STR];
+  u_char *last;
+
+  last = ngx_slprintf(buffer, buffer + NGX_MAX_ERROR_STR,
+                      "%s|%s%s\n",
+                      timestamp, info, message);
+  ngx_write_fd(data->file->fd, buffer, last - buffer);
+}
+
+static void
+ngx_http_groonga_query_logger_reopen(grn_ctx *ctx, void *user_data)
+{
+  ngx_http_groonga_query_logger_data_t *data = user_data;
+
+  GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_DESTINATION, " ",
+                "query log will be closed: <%.*s>",
+                (int)(data->path->len), data->path->data);
+  ngx_reopen_files((ngx_cycle_t *)ngx_cycle, -1);
+  GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_DESTINATION, " ",
+                "query log is opened: <%.*s>",
+                (int)(data->path->len), data->path->data);
+}
+
+static void
+ngx_http_groonga_query_logger_fin(grn_ctx *ctx, void *user_data)
+{
+  ngx_http_groonga_query_logger_data_t *data = user_data;
+
+  ngx_pfree(data->pool, data);
+}
+
+static grn_query_logger ngx_http_groonga_query_logger = {
+  GRN_QUERY_LOG_DEFAULT,
+  NULL,
+  ngx_http_groonga_query_logger_log,
+  ngx_http_groonga_query_logger_reopen,
+  ngx_http_groonga_query_logger_fin
+};
+
+static ngx_int_t
+ngx_http_groonga_context_init_query_logger(grn_ctx *context,
+                                           ngx_http_groonga_loc_conf_t *location_conf,
+                                           ngx_pool_t *pool,
+                                           ngx_log_t *log)
+{
+  ngx_http_groonga_query_logger_data_t *query_logger_data;
+
+  if (!location_conf->query_log_file) {
+    return NGX_OK;
+  }
+
+  query_logger_data = ngx_pcalloc(pool,
+                                  sizeof(ngx_http_groonga_query_logger_data_t));
+  if (!query_logger_data) {
+    ngx_log_error(NGX_LOG_ERR, log, 0,
+                  "http_groonga: failed to allocate memory for query logger");
+    return NGX_ERROR;
+  }
+
+  query_logger_data->pool = pool;
+  query_logger_data->file = location_conf->query_log_file;
+  query_logger_data->path = &(location_conf->query_log_path);
+  ngx_http_groonga_query_logger.user_data = query_logger_data;
+  grn_query_logger_set(context, &ngx_http_groonga_query_logger);
+
+  return NGX_OK;
+}
+
+static ngx_int_t
+ngx_http_groonga_context_init(grn_ctx *context,
+                              ngx_http_groonga_loc_conf_t *location_conf,
+                              ngx_pool_t *pool,
+                              ngx_log_t *log)
+{
+  ngx_int_t status;
+
+  grn_ctx_init(context, GRN_NO_FLAGS);
+
+  status = ngx_http_groonga_context_init_logger(context,
+                                                location_conf,
+                                                pool,
+                                                log);
+  if (status == NGX_ERROR) {
+    grn_ctx_fin(context);
+    return status;
+  }
+
+  status = ngx_http_groonga_context_init_query_logger(context,
+                                                      location_conf,
+                                                      pool,
+                                                      log);
+  if (status == NGX_ERROR) {
+    grn_ctx_fin(context);
+    return status;
+  }
+
+  if (location_conf->cache) {
+    grn_cache_current_set(context, location_conf->cache);
+  }
+
+  return status;
+}
+
 static void
 ngx_http_groonga_context_log_error(ngx_log_t *log, grn_ctx *context)
 {
@@ -127,10 +332,16 @@ ngx_http_groonga_handler_cleanup(void *user_data)
   ngx_http_groonga_handler_data_t *data = user_data;
   grn_ctx *context;
 
+  if (!data->initialized) {
+    return;
+  }
+
   context = &(data->context);
   GRN_OBJ_FIN(context, &(data->head));
   GRN_OBJ_FIN(context, &(data->body));
   GRN_OBJ_FIN(context, &(data->foot));
+  grn_logger_set(context, NULL);
+  grn_query_logger_set(context, NULL);
   grn_ctx_fin(context);
 }
 
@@ -209,7 +420,7 @@ ngx_http_groonga_extract_command_path(ngx_http_request_t *r,
   if (location_conf->base_path.len > 0) {
     if (command_path->len < location_conf->base_path.len) {
       ngx_log_error(NGX_LOG_WARN, r->connection->log, 0,
-                    "requestd URI is shorter than groonga_base_path: "
+                    "requested URI is shorter than groonga_base_path: "
                     "URI: <%V>, groonga_base_path: <%V>",
                     &(r->unparsed_uri), &(location_conf->base_path));
     } else if (strncmp((const char *)command_path->data,
@@ -266,7 +477,12 @@ ngx_http_groonga_handler_create_data(ngx_http_request_t *r,
   *data_return = data;
 
   context = &(data->context);
-  grn_ctx_init(context, GRN_NO_FLAGS);
+  rc = ngx_http_groonga_context_init(context, location_conf,
+                                     r->pool, r->connection->log);
+  if (rc != NGX_OK) {
+    return rc;
+  }
+  data->initialized = GRN_TRUE;
   GRN_TEXT_INIT(&(data->head), GRN_NO_FLAGS);
   GRN_TEXT_INIT(&(data->body), GRN_NO_FLAGS);
   GRN_TEXT_INIT(&(data->foot), GRN_NO_FLAGS);
@@ -331,26 +547,16 @@ ngx_http_groonga_handler_validate_post_command(ngx_http_request_t *r,
 }
 
 static ngx_int_t
-ngx_http_groonga_handler_process_body(ngx_http_request_t *r,
-                                      ngx_http_groonga_handler_data_t *data)
+ngx_http_groonga_send_lines(grn_ctx *context,
+                            ngx_http_request_t *r,
+                            u_char *current,
+                            u_char *last)
 {
   ngx_int_t rc;
 
-  grn_ctx *context;
+  u_char *line_start;
 
-  ngx_buf_t *body;
-  u_char *line_start, *current;
-
-  context = &(data->context);
-
-  body = r->request_body->buf;
-  if (!body) {
-    ngx_http_groonga_handler_set_content_type(r, "text/plain");
-    GRN_TEXT_PUTS(context, &(data->body), "must send load data as body");
-    return NGX_HTTP_BAD_REQUEST;
-  }
-
-  for (line_start = current = body->pos; current < body->last; current++) {
+  for (line_start = current; current < last; current++) {
     if (*current != '\n') {
       continue;
     }
@@ -375,6 +581,94 @@ ngx_http_groonga_handler_process_body(ngx_http_request_t *r,
   return NGX_OK;
 }
 
+static ngx_int_t
+ngx_http_groonga_join_request_body_chain(ngx_http_request_t *r,
+                                         ngx_chain_t *chain,
+                                         u_char **out_start,
+                                         u_char **out_end)
+{
+  ngx_int_t rc;
+
+  ngx_log_t *log = r->connection->log;
+
+  ngx_chain_t *current;
+  u_char *out;
+  size_t out_size;
+
+  u_char *out_cursor;
+  ngx_buf_t *buffer;
+  size_t buffer_size;
+
+  out_size = 0;
+  for (current = chain; current; current = current->next) {
+    out_size += ngx_buf_size(current->buf);
+  }
+  out = ngx_palloc(r->pool, out_size);
+  if (!out) {
+    ngx_log_error(NGX_LOG_ERR, log, 0,
+                  "http_groonga: failed to allocate memory for request body");
+    return NGX_ERROR;
+  }
+
+  out_cursor = out;
+  for (current = chain; current; current = current->next) {
+    buffer = current->buf;
+    buffer_size = ngx_buf_size(current->buf);
+
+    if (buffer->file) {
+      rc = ngx_read_file(buffer->file, out_cursor, buffer_size, 0);
+      if (rc < 0) {
+        ngx_log_error(NGX_LOG_ERR, log, 0,
+                      "http_groonga: failed to read a request body stored in a file");
+        return rc;
+      }
+    } else {
+      ngx_memcpy(out_cursor, buffer->pos, buffer_size);
+    }
+    out_cursor += buffer_size;
+  }
+
+  *out_start = out;
+  *out_end = out + out_size;
+
+  return NGX_OK;
+}
+
+static ngx_int_t
+ngx_http_groonga_handler_process_body(ngx_http_request_t *r,
+                                      ngx_http_groonga_handler_data_t *data)
+{
+  ngx_int_t rc;
+
+  grn_ctx *context;
+
+  ngx_buf_t *body;
+  u_char *body_data;
+  u_char *body_data_end;
+
+  context = &(data->context);
+
+  body = r->request_body->bufs->buf;
+  if (!body) {
+    ngx_http_groonga_handler_set_content_type(r, "text/plain");
+    GRN_TEXT_PUTS(context, &(data->body), "must send load data as body");
+    return NGX_HTTP_BAD_REQUEST;
+  }
+
+  rc = ngx_http_groonga_join_request_body_chain(r,
+                                                r->request_body->bufs,
+                                                &body_data,
+                                                &body_data_end);
+  if (rc != NGX_OK) {
+    return rc;
+  }
+
+  rc = ngx_http_groonga_send_lines(context, r, body_data, body_data_end);
+  ngx_pfree(r->pool, body_data);
+
+  return rc;
+}
+
 
 static ngx_int_t
 ngx_http_groonga_handler_process_load(ngx_http_request_t *r,
@@ -591,6 +885,114 @@ ngx_http_groonga_conf_set_groonga_slot(ngx_conf_t *cf, ngx_command_t *cmd,
   return NGX_CONF_OK;
 }
 
+static char *
+ngx_http_groonga_conf_set_log_path_slot(ngx_conf_t *cf, ngx_command_t *cmd,
+                                        void *conf)
+{
+  char *status;
+  ngx_http_groonga_loc_conf_t *groonga_location_conf = conf;
+
+  status = ngx_conf_set_str_slot(cf, cmd, conf);
+  if (status != NGX_CONF_OK) {
+    return status;
+  }
+
+  if (!groonga_location_conf->log_path.data) {
+    return NGX_CONF_OK;
+  }
+
+  if (strncmp((const char *)(groonga_location_conf->log_path.data),
+              "off",
+              groonga_location_conf->log_path.len) == 0) {
+    return NGX_CONF_OK;
+  }
+
+  groonga_location_conf->log_file =
+    ngx_conf_open_file(cf->cycle, &(groonga_location_conf->log_path));
+  if (!groonga_location_conf->log_file) {
+    ngx_log_error(NGX_LOG_ERR, cf->cycle->log, 0,
+                  "http_groonga: failed to open groonga log file: <%V>",
+                  &(groonga_location_conf->log_path));
+    return NGX_CONF_ERROR;
+  }
+
+  return NGX_CONF_OK;
+}
+
+static char *
+ngx_http_groonga_conf_set_log_level_slot(ngx_conf_t *cf, ngx_command_t *cmd,
+                                         void *conf)
+{
+  char *status = NGX_CONF_OK;
+  ngx_http_groonga_loc_conf_t *groonga_location_conf = conf;
+  char *value;
+
+  value = ngx_str_null_terminate(cf->cycle->pool,
+                                 ((ngx_str_t *)cf->args->elts) + 1);
+  if (strcasecmp(value, "none") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_NONE;
+  } else if (strcasecmp(value, "emergency") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_EMERG;
+  } else if (strcasecmp(value, "alert") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_ALERT;
+  } else if (strcasecmp(value, "critical") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_CRIT;
+  } else if (strcasecmp(value, "error") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_ERROR;
+  } else if (strcasecmp(value, "warning") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_WARNING;
+  } else if (strcasecmp(value, "notice") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_NOTICE;
+  } else if (strcasecmp(value, "info") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_INFO;
+  } else if (strcasecmp(value, "debug") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_DEBUG;
+  } else if (strcasecmp(value, "dump") == 0) {
+    groonga_location_conf->log_level = GRN_LOG_DUMP;
+  } else {
+    status = "must be one of 'none', 'emergency', 'alert', "
+      "'ciritical', 'error', 'warning', 'notice', 'info', 'debug' and 'dump'";
+  }
+  ngx_pfree(cf->cycle->pool, value);
+
+  return status;
+}
+
+static char *
+ngx_http_groonga_conf_set_query_log_path_slot(ngx_conf_t *cf,
+                                              ngx_command_t *cmd,
+                                              void *conf)
+{
+  char *status;
+  ngx_http_groonga_loc_conf_t *groonga_location_conf = conf;
+
+  status = ngx_conf_set_str_slot(cf, cmd, conf);
+  if (status != NGX_CONF_OK) {
+    return status;
+  }
+
+  if (!groonga_location_conf->query_log_path.data) {
+    return NGX_CONF_OK;
+  }
+
+  if (strncmp((const char *)(groonga_location_conf->query_log_path.data),
+              "off",
+              groonga_location_conf->query_log_path.len) == 0) {
+    return NGX_CONF_OK;
+  }
+
+  groonga_location_conf->query_log_file =
+    ngx_conf_open_file(cf->cycle, &(groonga_location_conf->query_log_path));
+  if (!groonga_location_conf->query_log_file) {
+    ngx_log_error(NGX_LOG_ERR, cf->cycle->log, 0,
+                  "http_groonga: failed to open groonga query log file: <%V>",
+                  &(groonga_location_conf->query_log_path));
+    return NGX_CONF_ERROR;
+  }
+
+  return NGX_CONF_OK;
+}
+
 static void *
 ngx_http_groonga_create_loc_conf(ngx_conf_t *cf)
 {
@@ -607,8 +1009,17 @@ ngx_http_groonga_create_loc_conf(ngx_conf_t *cf)
   conf->database_auto_create = NGX_CONF_UNSET;
   conf->base_path.data = NULL;
   conf->base_path.len = 0;
+  conf->log_path.data = NULL;
+  conf->log_path.len = 0;
+  conf->log_file = NULL;
+  conf->log_level = GRN_LOG_DEFAULT_LEVEL;
+  conf->query_log_path.data = NULL;
+  conf->query_log_path.len = 0;
+  conf->query_log_file = NULL;
+  conf->cache_limit = NGX_CONF_UNSET_SIZE;
   conf->config_file = NULL;
   conf->config_line = 0;
+  conf->cache = NULL;
 
   return conf;
 }
@@ -620,6 +1031,43 @@ ngx_http_groonga_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
   ngx_http_groonga_loc_conf_t *conf = child;
 
   ngx_conf_merge_str_value(conf->database_path, prev->database_path, NULL);
+  ngx_conf_merge_value(conf->database_auto_create,
+                       prev->database_auto_create,
+                       GRN_TRUE);
+  ngx_conf_merge_size_value(conf->cache_limit, prev->cache_limit,
+                            GRN_CACHE_DEFAULT_MAX_N_ENTRIES);
+
+#ifdef NGX_HTTP_GROONGA_LOG_PATH
+  {
+    ngx_str_t default_log_path;
+    default_log_path.data = (u_char *)NGX_HTTP_GROONGA_LOG_PATH;
+    default_log_path.len = strlen(NGX_HTTP_GROONGA_LOG_PATH);
+    conf->log_file = ngx_conf_open_file(cf->cycle, &default_log_path);
+    if (!conf->log_file) {
+      ngx_log_error(NGX_LOG_ERR, cf->cycle->log, 0,
+                    "http_groonga: "
+                    "failed to open the default groonga log file: <%V>",
+                    &default_log_path);
+      return NGX_CONF_ERROR;
+    }
+  }
+#else
+  conf->log_file = NULL;
+#endif
+
+  ngx_conf_merge_str_value(conf->query_log_path, prev->query_log_path,
+                           NGX_HTTP_GROONGA_QUERY_LOG_PATH);
+  if (!conf->query_log_file && conf->query_log_path.data && conf->enabled) {
+    conf->query_log_file = ngx_conf_open_file(cf->cycle,
+                                              &(conf->query_log_path));
+    if (!conf->query_log_file) {
+      ngx_log_error(NGX_LOG_ERR, cf->cycle->log, 0,
+                    "http_groonga: "
+                    "failed to open the default groonga query log file: <%V>",
+                    &(conf->query_log_path));
+      return NGX_CONF_ERROR;
+    }
+  }
 
   return NGX_CONF_OK;
 }
@@ -743,7 +1191,11 @@ ngx_http_groonga_open_database_callback(ngx_http_groonga_loc_conf_t *location_co
   grn_ctx *context;
 
   context = &(location_conf->context);
-  grn_ctx_init(context, GRN_NO_FLAGS);
+  data->rc = ngx_http_groonga_context_init(context, location_conf,
+                                           data->pool, data->log);
+  if (data->rc != NGX_OK) {
+    return;
+  }
 
   if (!location_conf->database_path.data) {
     ngx_log_error(NGX_LOG_EMERG, data->log, 0,
@@ -761,17 +1213,30 @@ ngx_http_groonga_open_database_callback(ngx_http_groonga_loc_conf_t *location_co
   }
 
   grn_db_open(context, location_conf->database_path_cstr);
-  if (context->rc == GRN_SUCCESS) {
-    return;
+  if (context->rc != GRN_SUCCESS) {
+    if (location_conf->database_auto_create) {
+      ngx_http_groonga_create_database(location_conf, data);
+    } else {
+      ngx_log_error(NGX_LOG_EMERG, data->log, 0,
+                    "failed to open groonga database: %s",
+                    context->errbuf);
+      data->rc = NGX_ERROR;
+      return;
+    }
   }
 
-  if (location_conf->database_auto_create) {
-    ngx_http_groonga_create_database(location_conf, data);
-  } else {
+  location_conf->cache = grn_cache_open(context);
+  if (!location_conf->cache) {
     ngx_log_error(NGX_LOG_EMERG, data->log, 0,
-                  "failed to open groonga database: %s",
+                  "failed to open groonga cache: %s",
                   context->errbuf);
     data->rc = NGX_ERROR;
+    return;
+  }
+  if (location_conf->cache_limit != NGX_CONF_UNSET_SIZE) {
+    grn_cache_set_max_n_entries(context,
+                                location_conf->cache,
+                                location_conf->cache_limit);
   }
 }
 
@@ -783,9 +1248,22 @@ ngx_http_groonga_close_database_callback(ngx_http_groonga_loc_conf_t *location_c
   grn_ctx *context;
 
   context = &(location_conf->context);
+  ngx_http_groonga_context_init_logger(context,
+                                       location_conf,
+                                       data->pool,
+                                       data->log);
+  ngx_http_groonga_context_init_query_logger(context,
+                                             location_conf,
+                                             data->pool,
+                                             data->log);
+  grn_cache_current_set(context, location_conf->cache);
 
   grn_obj_close(context, grn_ctx_db(context));
   ngx_http_groonga_context_log_error(data->log, context);
+
+  grn_cache_current_set(context, NULL);
+  grn_cache_close(context, location_conf->cache);
+
   grn_ctx_fin(context);
 }
 
@@ -801,8 +1279,11 @@ ngx_http_groonga_init_process(ngx_cycle_t *cycle)
     return NGX_ERROR;
   }
 
+  grn_set_segv_handler();
+
   http_conf =
     (ngx_http_conf_ctx_t *)ngx_get_conf(cycle->conf_ctx, ngx_http_module);
+
   data.log = cycle->log;
   data.pool = cycle->pool;
   data.rc = NGX_OK;
@@ -862,6 +1343,34 @@ static ngx_command_t ngx_http_groonga_commands[] = {
     offsetof(ngx_http_groonga_loc_conf_t, base_path),
     NULL },
 
+  { ngx_string("groonga_log_path"),
+    NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+    ngx_http_groonga_conf_set_log_path_slot,
+    NGX_HTTP_LOC_CONF_OFFSET,
+    offsetof(ngx_http_groonga_loc_conf_t, log_path),
+    NULL },
+
+  { ngx_string("groonga_log_level"),
+    NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+    ngx_http_groonga_conf_set_log_level_slot,
+    NGX_HTTP_LOC_CONF_OFFSET,
+    0,
+    NULL },
+
+  { ngx_string("groonga_query_log_path"),
+    NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+    ngx_http_groonga_conf_set_query_log_path_slot,
+    NGX_HTTP_LOC_CONF_OFFSET,
+    offsetof(ngx_http_groonga_loc_conf_t, query_log_path),
+    NULL },
+
+  { ngx_string("groonga_cache_limit"),
+    NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+    ngx_conf_set_size_slot,
+    NGX_HTTP_LOC_CONF_OFFSET,
+    offsetof(ngx_http_groonga_loc_conf_t, cache_limit),
+    NULL },
+
   ngx_null_command
 };
 
